import pytest
from unittest.mock import MagicMock, patch
from zero_sum_eval.managers.game_manager import GameManager

@pytest.fixture
def mock_game():
    """Create a mock game for testing"""
    game = MagicMock()
    game.is_over.side_effect = [False, True]  # Game ends after first move
    
    # Create a proper mock action
    mock_action = MagicMock()
    mock_action.name = "test_action"  # Set name as attribute, not as mock
    mock_action.player = MagicMock(
        id="player1",
        llm_model="mock_model",
        module_dict={
            "test_action": MagicMock(  # Match the action name
                return_value=MagicMock(
                    items=lambda: [("output", "test")]
                )
            )
        }
    )
    
    game.get_next_action.return_value = mock_action
    game.player_inputs.return_value = {}
    game.display.return_value = "game state"
    game.export.return_value = {"state": "test"}
    return game

@pytest.fixture
def mock_config():
    """Create a mock configuration for testing"""
    return {
        "manager": {
            "args": {
                "max_rounds": 10,
                "max_games": 5,
                "max_matches": 3,
                "max_player_attempts": 3
            }
        },
        "logging": {
            "output_dir": "/tmp/logs"
        }
    }

def test_game_manager_initialization(mock_config):
    """Test GameManager initialization"""
    manager = GameManager(mock_config)
    assert manager.config == mock_config
    assert manager.max_rounds == mock_config["manager"]["args"]["max_rounds"]
    # Remove turn_timeout check since it's not part of GameManager

@patch('dspy.context')
@patch('jsonlines.Writer')
def test_game_manager_start(mock_writer, mock_context, mock_config, mock_game):
    """Test game manager start functionality"""
    manager = GameManager(mock_config)
    manager.game = mock_game
    
    result = manager.start(mock_game)
    
    # Verify game was played correctly
    assert result == mock_game
    assert mock_game.update_game.called
    assert mock_game.get_next_action.called
    assert mock_game.player_inputs.called
    assert mock_game.export.called

@patch('dspy.context')
@patch('jsonlines.Writer')
def test_game_manager_turn_handling(mock_writer, mock_context, mock_config, mock_game):
    """Test game manager turn handling"""
    manager = GameManager(mock_config)
    manager.game = mock_game
    
    # Setup mock action
    mock_action = MagicMock()
    mock_action.name = "test_action"  # Set as attribute
    mock_action.player = MagicMock(
        id="player1",
        player_key="test_player",
        llm_model="mock_model",
        module_dict={
            "test_action": MagicMock(  # Match the action name
                return_value=MagicMock(
                    items=lambda: [("output", "test")]
                )
            )
        }
    )
    mock_game.get_next_action.return_value = mock_action
    
    # Test turn execution
    result = manager.start(mock_game)
    
    # Verify turn was handled correctly
    next_action = result.get_next_action()
    assert next_action == mock_action
    assert next_action.name == "test_action"
    assert next_action.player.player_key == "test_player"

@patch('dspy.context')
@patch('jsonlines.Writer')
def test_game_manager_logging(mock_writer, mock_context, mock_config, mock_game):
    """Test game manager logging functionality"""
    manager = GameManager(mock_config)
    manager.game = mock_game
    
    manager.start(mock_game)
    
    # Verify logging occurred
    assert mock_game.export.called
    assert mock_game.display.called

@patch('dspy.context')
@patch('jsonlines.Writer')
def test_game_manager_error_handling(mock_writer, mock_context, mock_config, mock_game):
    """Test game manager error handling"""
    manager = GameManager(mock_config)
    manager.game = mock_game
    
    # Setup error condition
    mock_game.update_game.side_effect = Exception("Test error")
    
    # Test error handling
    with pytest.raises(Exception):
        manager.start(mock_game) 